// This file is part of ICU4X. For terms of use, please see the file
// called LICENSE at the top level of the ICU4X source tree
// (online at: https://github.com/unicode-org/icu4x/blob/main/LICENSE ).

#[cfg(feature = "alloc")]
mod enum_keyword;
mod struct_keyword;

#[doc(inline)]
#[cfg(feature = "alloc")]
pub use enum_keyword::enum_keyword;
#[doc(inline)]
pub use struct_keyword::struct_keyword;
